/* ------------------------------------------------------------- */
/* - STM32 I2C LCD module controller                             */
/*                                  released by fumi2kick        */
/*                                  http://www.fumi2kick.com/    */
/*                                  rerofumi <rero2@yuumu.org>   */
/* ------------------------------------------------------------- */
/* update                                                        */
/*   Mar.20.2011 : first release                                 */
/* ------------------------------------------------------------- */
/* memo                                                          */
/*                                                               */
/* ------------------------------------------------------------- */

/* ----------------------------------- */
/* --- include                         */
/* ----------------------------------- */
#include "stm32f10x_conf.h"

#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"
#include "freertos/task.h"

#include "fm_i2c_lcd.h"


/* ----------------------------------- */
/* --- define                          */
/* ----------------------------------- */

/* IRQ message type */
typedef unsigned int FmI2cIrqEvent;

/* IRQ message */
#define FM_I2CIRQ_NULL               0x0000
#define FM_I2CIRQ_DONE               0x0001
#define FM_I2CIRQ_QUIT               0x0002
#define FM_I2CIRQ_ERROR_MASK         0x8000
#define FM_I2CIRQ_ERROR_SOMETHING    0x8000
#define FM_I2CIRQ_ERROR_TIMEOUT      0x8001
#define FM_I2CIRQ_ERROR_NACK         0x8002
#define FM_I2CIRQ_ERROR_BUS          0x8003

/* SEND RESULT */
#define FM_I2CSEND_DONE              0
#define FM_I2CSEND_ERROR             1
#define FM_I2CSEND_TIMEOUT           2

/* master send state */
enum {
  I2C_STATE_START = 0,
  I2C_STATE_ADDRESS_DONE,
  I2C_STATE_DATA_1_DONE,
  I2C_STATE_DATA_2_DONE,
  I2C_STATE_STOP,
} MASTER_SEND_STATE;

/* Command request queue SIZE */
#define  CMDREQ_SIZE  16

/* HEX char */
const unsigned char cHexCode[16] = {
  '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
};

/* ----------------------------------- */
/* --- local function prototyping      */
/* ----------------------------------- */

/* --- module reset */
void module_reset(void);

/* --- I2C send one command */
int send_cmd(char rs, unsigned char data);


/* ----------------------------------- */
/* --- global scope                   */
/* ----------------------------------- */

/* --- I2C Send Request result queue */
xQueueHandle  queueI2cIrq;

/* --- display command queue */
xQueueHandle  queueCmdReq;

/* --- I2C port */
I2C_TypeDef *pI2cPort;

/* --- I2C IRQ state */
int stateI2cSend;

/* --- send data */
unsigned char bSendData[2];

/* --- LCD Manager Task statement */
int stateI2cLcdManager;


/* ------------------------------------------------------------- */
/* -----  interface                                              */
/* ------------------------------------------------------------- */

/* ------------------------------- */
/* --- initialize */
int fmI2C_Init(int port)
{
  I2C_InitTypeDef initI2c;
  NVIC_InitTypeDef NVIC_InitStructure;

  stateI2cLcdManager = STATE_I2CLCD_NULL;

  /* IRQ message queue (4byte x 2message) */
  queueI2cIrq = xQueueCreate(2, sizeof(FmI2cIrqEvent));
  /* Command Request queue */
  queueCmdReq = xQueueCreate(CMDREQ_SIZE, sizeof(uint16_t));

  /* I2C initialize */
  if (port == 1) {
    pI2cPort = I2C2;
    /* Interrupt initilize */
    NVIC_InitStructure.NVIC_IRQChannel = I2C2_EV_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 13;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);
    NVIC_InitStructure.NVIC_IRQChannel = I2C2_ER_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 14;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);
  }
  else {
    pI2cPort = I2C1;
    /* Interrupt initilize */
    NVIC_InitStructure.NVIC_IRQChannel = I2C1_EV_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 13;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);
    NVIC_InitStructure.NVIC_IRQChannel = I2C1_ER_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 14;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);
  }

  I2C_Cmd(pI2cPort, DISABLE);
  I2C_StructInit(&initI2c);
  initI2c.I2C_Ack = I2C_Ack_Enable;
  initI2c.I2C_ClockSpeed = 400000;
  I2C_Init(pI2cPort, &initI2c);
  I2C_ITConfig(pI2cPort, I2C_IT_EVT, ENABLE);
  //I2C_ITConfig(pI2cPort, I2C_IT_BUF, ENABLE);
  I2C_ITConfig(pI2cPort, I2C_IT_ERR, ENABLE);
  I2C_ClearITPendingBit(pI2cPort, I2C_IT_TIMEOUT|I2C_IT_PECERR|I2C_IT_OVR|I2C_IT_AF|I2C_IT_ARLO|I2C_IT_BERR);
  I2C_Cmd(pI2cPort, ENABLE);

  stateI2cLcdManager = STATE_I2CLCD_NONE_INITALIZE;

  return 0;
}


/* ------------------------------- */
/* --- manager state */
int fmI2C_LCD_State(void)
{
  int st;
  st = stateI2cLcdManager;
  return st;
}


/* ------------------------------- */
/* --- Clear Screen */
void fmI2C_LCD_ClearScreen(void)
{
  uint16_t cmd;

  cmd = (30 << 8) | 0b00000001;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
}


/* ------------------------------- */
/* --- Clear Screen */
void fmI2C_LCD_Contrast(int cont)
{
  int val;
  unsigned char reg;
  uint16_t cmd;

  val = cont;
  if (val < 0) {
    val = 0;
  }
  if (val > 255) {
    val = 255;
  }
  reg = (unsigned char)(val >> 2);

  /* req */
  cmd = 0b00111000;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
  cmd = 0b00111001;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
  cmd = 0b01110000 | (reg & 0b00001111);
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
  cmd = 0b01010100 | ((reg & 0b00110000) >> 4);
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
  cmd = 0b00111000;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
}


/* ------------------------------- */
/* --- locate */
void fmI2C_LCD_Position(int x, int y)
{
  uint16_t cmd;

  cmd = 0b10000000 + ((y & 1) * 0x40) + x;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
}


/* ------------------------------- */
/* --- display 1byte char */
void fmI2C_LCD_PutChar(const unsigned char cCode)
{
  uint16_t cmd;

  cmd = 0x8000 + cCode;
  xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
}


/* ------------------------------- */
/* --- display line */
void fmI2C_LCD_Puts(const unsigned char *cString, const int size)
{
  uint16_t cmd;
  int i;

  for(i=0; i<size; i++) {
    if (cString[i] == 0) break;
    /* put */
    cmd = 0x8000 + cString[i];
    xQueueSend(queueCmdReq, (void*)&cmd, portMAX_DELAY);
  }
}


/* ------------------------------- */
/* --- display HEX value */
void fmI2C_LCD_DisplayHex_8(const uint8_t val)
{
  unsigned char c_low;

  c_low = cHexCode[((val & 0xf0) >> 4)];
  fmI2C_LCD_PutChar(c_low);
  c_low = cHexCode[(val & 0x0f)];
  fmI2C_LCD_PutChar(c_low);
}

void fmI2C_LCD_DisplayHex_16(const uint16_t val)
{
  unsigned char c_low;

  c_low = (val & 0xff00) >> 8;
  fmI2C_LCD_DisplayHex_8(c_low);
  c_low = val & 0x00ff;
  fmI2C_LCD_DisplayHex_8(c_low);
}

void fmI2C_LCD_DisplayHex_32(const uint32_t val)
{
  uint16_t c_low;

  c_low = (val & 0xffff0000) >> 16;
  fmI2C_LCD_DisplayHex_16(c_low);
  c_low = val & 0x0000ffff;
  fmI2C_LCD_DisplayHex_16(c_low);
}



/* ------------------------------------------------------------- */
/* -----  task                                                   */
/* ------------------------------------------------------------- */

void vI2C_LCD_TASK(void *pcParam)
{
  int initialize_result;
  uint16_t cmd;
  int wait;
  unsigned char c_rs, c_data;

  stateI2cLcdManager = STATE_I2CLCD_INITALIZING;
  GPIO_ResetBits(GPIOB, 0xe000);

  /* --- reset */
  module_reset();

  initialize_result = 0;

  /* --- Initialize (function) */
  vTaskDelay(50 / portTICK_RATE_MS);
  initialize_result |= send_cmd(0, 0b00111000);
  vTaskDelay(1 / portTICK_RATE_MS);
  initialize_result |= send_cmd(0, 0b00111001);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* OSC frequency */
  initialize_result |= send_cmd(0, 0b00010100);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* contrast */
  initialize_result |= send_cmd(0, 0b01111000);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* icon */
  initialize_result |= send_cmd(0, 0b01010110);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* follower */
  initialize_result |= send_cmd(0, 0b01101100);
  vTaskDelay(200 / portTICK_RATE_MS);
  /* normal function */
  initialize_result |= send_cmd(0, 0b00111000);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* display on */
  initialize_result |= send_cmd(0, 0b00001100);
  vTaskDelay(1 / portTICK_RATE_MS);
  /* clear */
  initialize_result |= send_cmd(0, 0b00000001);
  vTaskDelay(2 / portTICK_RATE_MS);

  /* --- umm... got error */
  if (FM_I2CSEND_DONE != initialize_result) {
    vTaskSuspend(NULL);
  }

  stateI2cLcdManager = STATE_I2CLCD_READY;

  while(1) {
    xQueueReceive(queueCmdReq, (void*)&cmd, portMAX_DELAY);
    /* prepare */
    stateI2cLcdManager = STATE_I2CLCD_WORKING;
    wait = (cmd & 0x7f00) >> 8;
    c_rs = ((cmd & 0x8000) != 0);
    c_data = cmd & 0xff;
    /* send */
    send_cmd(c_rs, c_data);
    /* wait */
    if (wait > 0) {
      vTaskDelay(wait / portTICK_RATE_MS);
    }
    /* done */
    stateI2cLcdManager = STATE_I2CLCD_READY;
  }
}


/* ------------------------------------------------------------- */
/* -----  IRQ                                                    */
/* ------------------------------------------------------------- */

/* ------------------------------------- */
/* --- I2C send state                    */
/* ------------------------------------- */
void fmI2C_IRQ_EV(void)
{
  static FmI2cIrqEvent eventCode;
  static portBASE_TYPE xHigherPriorityTaskWoken;

  eventCode = FM_I2CIRQ_NULL;

  /* --- I2C send state */
  if (I2C_STATE_START == stateI2cSend) {
    if (I2C_GetITStatus(pI2cPort, I2C_IT_SB) != RESET) {
      /* I2C master send address */
      I2C_Send7bitAddress(pI2cPort, (FM_I2CLCD_ADDRESS<<1), I2C_Direction_Transmitter); 
      /* next */
      stateI2cSend = I2C_STATE_ADDRESS_DONE;
    }
  }
  else if (I2C_STATE_ADDRESS_DONE == stateI2cSend) {
    if (I2C_CheckEvent(pI2cPort, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED) == SUCCESS) {
      /* send rs */
      I2C_SendData(pI2cPort, bSendData[0]);
      /* next */
      stateI2cSend = I2C_STATE_DATA_1_DONE;
    }
  }
  else if (I2C_STATE_DATA_1_DONE == stateI2cSend) {
    if (I2C_CheckEvent(pI2cPort, I2C_EVENT_MASTER_BYTE_TRANSMITTED) == SUCCESS) {
      /* send rs */
      I2C_SendData(pI2cPort, bSendData[1]);
      /* next */
      stateI2cSend = I2C_STATE_DATA_2_DONE;
    }
  }
  else if (I2C_STATE_DATA_2_DONE == stateI2cSend) {
    if (I2C_CheckEvent(pI2cPort, I2C_EVENT_MASTER_BYTE_TRANSMITTED) == SUCCESS) {
      /* stop condition */
      I2C_GenerateSTOP(pI2cPort, ENABLE);
      stateI2cSend = I2C_STATE_STOP;
      /* done message */
      eventCode = FM_I2CIRQ_DONE;
      xHigherPriorityTaskWoken = pdFALSE;
      xQueueSendFromISR(queueI2cIrq, (void*)&eventCode, &xHigherPriorityTaskWoken);
      portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
    }
  }
}

void fmI2C_IRQ_ER(void)
{
  FmI2cIrqEvent errorCode;
  portBASE_TYPE xHigherPriorityTaskWoken;

  errorCode = FM_I2CIRQ_ERROR_SOMETHING;

  /* --- ŃXe[^Xǂݎ */
  if (I2C_GetITStatus(pI2cPort, I2C_IT_SMBALERT) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_SMBALERT);
    errorCode = FM_I2CIRQ_ERROR_BUS;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_TIMEOUT) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_TIMEOUT);
    errorCode = FM_I2CIRQ_ERROR_TIMEOUT;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_PECERR) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_PECERR);
    errorCode = FM_I2CIRQ_ERROR_NACK;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_OVR) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_OVR);
    errorCode = FM_I2CIRQ_ERROR_BUS;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_AF) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_AF);
    errorCode = FM_I2CIRQ_ERROR_NACK;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_ARLO) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_ARLO);
    errorCode = FM_I2CIRQ_ERROR_BUS;
  }
  if (I2C_GetITStatus(pI2cPort, I2C_IT_BERR) != RESET) {
    I2C_ClearITPendingBit(pI2cPort, I2C_IT_BERR);
    errorCode = FM_I2CIRQ_ERROR_BUS;
  }

  /* send message */
  xHigherPriorityTaskWoken = pdFALSE;
  if (errorCode != FM_I2CIRQ_ERROR_SOMETHING) {
    xQueueSendToBackFromISR(queueI2cIrq, (void*)&errorCode, &xHigherPriorityTaskWoken);
    // context switch : Set PendSV bit
    portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
  }
}


/* ------------------------------------------------------------- */
/* -----  local function                                         */
/* ------------------------------------------------------------- */

void module_reset(void)
{
#if (FM_I2CLCD_SW_USERESET == 1)
  GPIO_SetBits(FM_I2CLCD_RESETPORT, FM_I2CLCD_RESETPIN);
  vTaskDelay(1 / portTICK_RATE_MS);
  GPIO_ResetBits(FM_I2CLCD_RESETPORT, FM_I2CLCD_RESETPIN);
  vTaskDelay(1 / portTICK_RATE_MS);
  GPIO_SetBits(FM_I2CLCD_RESETPORT, FM_I2CLCD_RESETPIN);
#endif  
}


int send_cmd(char rs, unsigned char data)
{
  FmI2cIrqEvent eventCode;
  int result;

  result = FM_I2CSEND_ERROR;

  /* --- prepare */
  bSendData[0] = (rs & 1) << 6;
  bSendData[1] = data;
  stateI2cSend = I2C_STATE_START;

  /* --- start */
  I2C_GenerateSTART(pI2cPort, ENABLE);

  /* --- sync */
  if (xQueueReceive(queueI2cIrq, (void*)&eventCode, (50/portTICK_RATE_MS)) == SUCCESS) {
    if ((eventCode & FM_I2CIRQ_ERROR_MASK) != 0) {
      I2C_GenerateSTOP(pI2cPort, ENABLE);
      result = FM_I2CSEND_ERROR;
    }
    else {
      result = FM_I2CSEND_DONE;
    }
  }
  else {
    I2C_GenerateSTOP(pI2cPort, ENABLE);
    result = FM_I2CSEND_TIMEOUT;
  }

  return result;
}




